package net.sergeych.bipack

import kotlinx.serialization.SerialInfo

/**
 * If this annotation is presented in some @Serializable class defition, its instances
 * will be serialized with leadinf number of fields. This allows to extend class later
 * providing new parameters __to the end of the class__ and _with default values__.
 *
 * Whe deserializing such instances from previous version binaries, the new parameters
 * will get default values.
 *
 * Serialized data of classes not market as ExtendableFormat could not be changed without
 * breaking compatibility with existing serialized data.
 */
@Target(AnnotationTarget.CLASS)
@SerialInfo
annotation class ExtendableFormat


/**
 * Serializable classes annotated as Framed will have leading checked mark as CRC32
 * of its name (uses `@SerialName` internally). On deserializing, if frame will not
 * match the expected name, the [InvalidFrameException] will be thrown.
 */
@SerialInfo
annotation class Framed

/**
 * Allow to CRC-protect structures (we suppose to use it with classes only). After the
 * data block its CRC32 will be written and checked. It is memory-wise: it calculates CRC
 * on the fly without buffering the data. If used with [Framed] and [ExtendableFormat] the extra
 * data is protected too.
 *
 * __Common pitfalls__. When unpacking corrupted data protected this way, the not only [InvalidFrameCRCException]
 * can be thrown. Actually, most often you will see [DataSource.EndOfData] exception
 */
@SerialInfo
annotation class CrcProtected

/**
 * Allow marking data fields as being serialized as usnsigned (applyable also to signed fields lite Int, Long and
 * Short, if you are shure they will not be negative). As unsigned types are not cully supported by kotlinx.serialization
 * it is the conly way to tell the serialized to use more compact unsigned variable length encoding.
 */
@SerialInfo
@Target(AnnotationTarget.FIELD,AnnotationTarget.PROPERTY)
annotation class Unsigned

open class InvalidFrameException(reason: String) : Exception(reason)
class InvalidFrameHeaderException(reason: String = "Frame header does not match") : InvalidFrameException(reason)
class InvalidFrameCRCException : InvalidFrameException("Checksum CRC32 failed")



