package bipack

import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import net.sergeych.bintools.toDump
import net.sergeych.bipack.*
import kotlin.experimental.xor
import kotlin.test.*

@Serializable
data class Foobar1N(val bar: Int, val foo: Int = 117)

@Serializable
@ExtendableFormat
data class Foobar1(val bar: Int, val foo: Int = 117)

@Serializable
@ExtendableFormat
data class Foobar2(val bar: Int, val foo: Int, val other: Int = -1)

@Serializable
@Framed
@ExtendableFormat
data class FoobarF1(val bar: Int, val foo: Int = 117)

@Serializable
@Framed
@ExtendableFormat
@SerialName("bipack.FoobarF1")
data class FoobarF2(val bar: Int, val foo: Int, val other: Int = -1)

@Serializable
@Framed
@ExtendableFormat
data class FoobarF3(val bar: Int, val foo: Int, val other: Int = -1)

@Serializable
@Framed
@ExtendableFormat
@CrcProtected()
data class FoobarFP1(val bar: Int, val foo: Int, val other: Int = -1)

class BipackEncoderTest {

    @Serializable
    data class FoobarSize(val i: Int)

    @Test
    fun testSize() {
        println(BipackEncoder.encode(17).toDump())
        assertEquals(1, BipackEncoder.encode(17).size)
//        assertEquals(1, BipackEncoder.encode(FoobarSize(17)).size)
    }


    @Test
    fun encodeSimple() {
        val a = Foobar1(42)//, "bum")
        println(BipackEncoder.encode(a).toDump())
        val b = BipackDecoder.decode<Foobar1>(BipackEncoder.encode(a))
        assertEquals(a, b)
        val c = BipackDecoder.decode<Foobar2>(BipackEncoder.encode(a))
        assertEquals(-1, c.other)
        assertEquals(a.bar, c.bar)
    }

    @Test
    fun encodeFramed() {
        val a = FoobarF1(42)//, "bum")
        println(BipackEncoder.encode(a).toDump())
        val b = BipackDecoder.decode<FoobarF1>(BipackEncoder.encode(a))
        assertEquals(a, b)
        val c = BipackDecoder.decode<FoobarF2>(BipackEncoder.encode(a))
        assertEquals(-1, c.other)
        assertEquals(a.bar, c.bar)
        assertFailsWith(InvalidFrameException::class) {
            BipackDecoder.decode<FoobarF3>(BipackEncoder.encode(a))
        }
    }

    @Test
    fun encodeNonExtendable() {
        val a = Foobar1N(42)//, "bum")
        println(BipackEncoder.encode(a).toDump())
        val b = BipackDecoder.decode<Foobar1N>(BipackEncoder.encode(a))
        assertEquals(a, b)
    }

    @Test
    fun encodeProtectedFrame() {
        val a = FoobarFP1(42, 117)//, "bum")
        println(BipackEncoder.encode(a).toDump())
        val b = BipackDecoder.decode<FoobarFP1>(BipackEncoder.encode(a))
        assertEquals(a, b)
        val bad = BipackEncoder.encode(a)
        bad[6] = bad[6] xor 0x20
        println(bad.toDump())
        assertFailsWith(InvalidFrameCRCException::class) {
            BipackDecoder.decode<FoobarFP1>(bad)
        }
    }

    @Serializable
    data class FBU(
        @Unsigned
        val u: UInt,
        @Unsigned
        val i: Int,
        val k: UInt = 3u,
    )

    @Test
    fun testByteArray() {
//        val z = Foobar1(42)//, "bum")
//        println(BipackEncoder.encode(z).toDump())

        val x = byteArrayOf(1, 2, 3)
        var d = BipackEncoder.encode(x)
        println(d.toDump())
        assertEquals(0x0c, d[0])

        d = BipackEncoder.encode("123")
        println(d.toDump())
        assertEquals(0x0c, d[0])

        val f = FBU(3u, 3)
        d = BipackEncoder.encode(f)
        println(d.toDump())
        assertEquals(0x0c, d[0])
        assertEquals(0x0c, d[1])
        // Signed encoding despite UInt type:
        assertEquals(0x18, d[2])

        assertEquals(f, BipackDecoder.decode(d))
    }

    @Serializable
    data class Types1(
        val i: Int = 7,
        val f: Float = 1f/3f,
        val d: Double = 1.0/3.0,
        val b: Boolean = true,
        val s: String = "жёпа",
        val ba: ByteArray = byteArrayOf(1,2,3),
        val ch: Char = 'Ы',
    ) {
        override fun equals(other: Any?): Boolean {
            if (this === other) return true
            if (other == null || this::class != other::class) return false

            other as Types1

            if (i != other.i) return false
            if (f != other.f) return false
            if (d != other.d) return false
            if (b != other.b) return false
            if (s != other.s) return false
            if (!ba.contentEquals(other.ba)) return false
            if (ch != other.ch) return false

            return true
        }

        override fun hashCode(): Int {
            var result = i
            result = 31 * result + f.hashCode()
            result = 31 * result + d.hashCode()
            result = 31 * result + b.hashCode()
            result = 31 * result + s.hashCode()
            result = 31 * result + ba.contentHashCode()
            result = 31 * result + ch.hashCode()
            return result
        }
    }

    @Test
    fun testTypes() {
//        val t1 = Types1(10, 1.0f / 10.0f, 1.0 / 10.0, true, "жесть", byteArrayOf(1, 2, 3), 'Ы')
        val t1 = Types1(f=17f/7f)
        val d = BipackEncoder.encode(t1)
//        println(d.toDump())
//        println(t1)
//        println(d.decodeFromBipack<Types1>())
        assertEquals(t1, d.decodeFromBipack())
    }
}